import pandas as pd
import matplotlib.pyplot as plt
import numpy as np
from matplotlib import gridspec
from SoftwareVersion import software_version

def extract_stable_data(datafile, measurement_set, reference_signal_names, extra_signal_names):
    datapoint = {}
    # df is a pandas data frame
    df = pd.read_csv(datafile)

    # extract phase mean and variance for stable measurements (don't ask what loc means)
    signal0 = df.loc[(df['EQUILIBRIUM_INDICATOR'] == 31) & (df['SET_NAME'] == measurement_set),
                     reference_signal_names[0]]
    if signal0.size == 0:
        return None
    datapoint['signal0_mean']=signal0.mean()
    datapoint['signal0_var']=signal0.var()

    signal1 = df.loc[(df['EQUILIBRIUM_INDICATOR'] == 31) & (df['SET_NAME'] == measurement_set),
                     reference_signal_names[1]]
    datapoint['signal1_mean'] = signal1.mean()
    datapoint['signal1_var'] = signal1.var()

    temperatures = df.loc[(df['EQUILIBRIUM_INDICATOR'] == 31) & (df['SET_NAME'] == measurement_set), 'READBACK_TEMPERATURE']
    datapoint['temperature_mean'] = temperatures.mean()
    datapoint['temperature_var'] = temperatures.var()

    humidities = df.loc[(df['EQUILIBRIUM_INDICATOR'] == 31) & (df['SET_NAME'] == measurement_set), 'READBACK_HUMIDITY']
    datapoint['humidity_mean'] = humidities.mean()
    datapoint['humidity_var'] = humidities.var()

    for extra_signal_name in extra_signal_names:
        extra_signal_values = df.loc[(df['EQUILIBRIUM_INDICATOR'] == 31) & (df['SET_NAME'] == measurement_set), extra_signal_name]
        datapoint[extra_signal_name] = extra_signal_values.mean()

    return datapoint


# sweep_type is either 'temperature' or 'humidity'
def plot_sweep(temperatures, humidities, basename, sweep_type, measurement_sets, reference_signal_names,
               analysis_config):
    set_data = {}
    derivatives = {}
    for measurement_set in measurement_sets:
        set_data[measurement_set] = {'signal0_means': [], 'signal0_vars': [], 'signal1_means': [], 'signal1_vars': [],
                                     'x_data': []}
        for external_signal_name in analysis_config['extra_signal_names']:
            set_data[measurement_set][external_signal_name] = []
        derivatives[measurement_set] = {'signal0_deltas': [], 'signal1_deltas': [], 'x_deltas': []}
        
    for temp, hum in zip(temperatures, humidities):
        datafile = basename+'_'+str(temp)+'deg_'+str(hum)+'rh.csv'
        print(datafile)
        for measurement_set in measurement_sets:
            datapoint = extract_stable_data(datafile, measurement_set, reference_signal_names,
                                            analysis_config['extra_signal_names'])
            if datapoint is None:
                continue

            data = set_data[measurement_set]
            if sweep_type == 'temperature':
                data['x_data'].append(datapoint['temperature_mean'])
            elif sweep_type == 'humidity':
                data['x_data'].append(datapoint['humidity_mean'])
            else:
                raise Exception('Unknown sweep_type:'+str(sweep_type))

            data['signal0_means'].append(datapoint['signal0_mean'])
            data['signal0_vars'].append(datapoint['signal0_var'])
            data['signal1_means'].append(datapoint['signal1_mean'])
            data['signal1_vars'].append(datapoint['signal1_var'])

            for external_signal_name in analysis_config['extra_signal_names']:
                data[external_signal_name].append(datapoint[external_signal_name])

    for measurement_set in measurement_sets:
        data = set_data[measurement_set]
        if analysis_config['normalise'][0]:
            data['signal0_means'] -= data['signal0_means'][0]
        if analysis_config['normalise'][1]:
            data['signal1_means'] -= data['signal1_means'][0]

    fig = plt.figure(figsize=(10, 15))
    gs = gridspec.GridSpec(nrows=6, ncols=1, hspace=0)#, width_ratios=[3, 1], height_ratios=[3, 1])
    
    upper_text_block = '\n'.join(['DUT name: '+analysis_config['dut_name'],
                                  'SW version: '+software_version(), 'Date: '+analysis_config['time_string']])

    
    fig.text(0.1, 0.88, upper_text_block)
    
    ax1 = fig.add_subplot(gs[1, 0])
    ax2 = fig.add_subplot(gs[2, 0], sharex=ax1)
    ax4 = fig.add_subplot(gs[4, 0])
    ax5 = fig.add_subplot(gs[5, 0], sharex=ax4)
    
    ax1.tick_params(bottom=True, top=True, direction='in')
    ax2.tick_params(bottom=True, top=True, direction='inout')
    ax4.tick_params(bottom=True, top=True, direction='in')
    ax5.tick_params(bottom=True, top=True, direction='inout')

    if sweep_type == 'temperature':
        fig.suptitle(basename + ': Temperature sweep at ' + str(humidities[0]) + ' % r.h.')
        ax2.set_xlabel('temperature [deg C]')
        ax5.set_xlabel('temperature [deg C]')
        denominator_name = '/ deg C'
    elif sweep_type == 'humidity':
        fig.suptitle(basename + ': Humidity sweep at ' + str(temperatures[0]) + ' deg C')
        ax2.set_xlabel('relative humidity [%]')
        ax5.set_xlabel('relative humidity [%]')
        denominator_name = '/ % r.h.'
    else:
        raise Exception('Unknown sweep_type:'+str(sweep_type))

    for measurement_set in measurement_sets:
        ax1.errorbar(set_data[measurement_set]['x_data'],
                     set_data[measurement_set]['signal0_means'], set_data[measurement_set]['signal0_vars'],
                     label=measurement_set, marker='+', linewidth=0)
    ax1.set_ylabel(reference_signal_names[0])

    for measurement_set in measurement_sets:
        ax2.errorbar(set_data[measurement_set]['x_data'],
                     set_data[measurement_set]['signal1_means'], set_data[measurement_set]['signal1_vars'],
                     label=measurement_set, marker='+', linewidth=0)
    ax2.set_ylabel(reference_signal_names[1])

    # Remove the highest tick label from the lower subplot. It probably overlaps with the lowest one of the upper plot.
    yticks = ax2.yaxis.get_major_ticks()
    yticks[-1].set_visible(False)
    yticks[-2].label1.set_visible(False)  # don't ask me why we have to set the last two tick marks to invisible
    ax2.legend(loc='center right', bbox_to_anchor=(1.5, 1.0))

    #######################################################################################################################
    # Derrivatives
    #######################################################################################################################
    for measurement_set in measurement_sets:
        data = set_data[measurement_set]
        deriv_data = derivatives[measurement_set]
        for i in range(len(data['signal0_means'])-1):
            ylabel = '$\\Delta$ '+reference_signal_names[0]+' '+denominator_name
            deriv_data['signal0_deltas'].append((data['signal0_means'][i+1]-data['signal0_means'][i])/
                                                (data['x_data'][i+1]-data['x_data'][i]))
            deriv_data['signal1_deltas'].append((data['signal1_means'][i+1]-data['signal1_means'][i])/
                                                (data['x_data'][i+1]+data['x_data'][i]))
            deriv_data['x_deltas'].append((data['x_data'][i+1]+data['x_data'][i])/2)

    ax4.set_ylabel('$\\Delta$ '+reference_signal_names[0]+' '+denominator_name)
    ax5.set_ylabel('$\\Delta$ '+reference_signal_names[1]+' '+denominator_name)

    #######################################################################################################################
    # RF cable post analysis: Normalise to cable length and convert phase to time
    #######################################################################################################################
    
    if analysis_config['type'] == 'rf_cable':
        for measurement_set in measurement_sets:
            data = set_data[measurement_set]
            deriv_data = derivatives[measurement_set]
            phase_to_time = 1e15 / (analysis_config['cable_length'] * data['RF_FREQUENCY'][0] * 360.)
            deriv_data['signal0_deltas'][:] = [x * phase_to_time for x in deriv_data['signal0_deltas']]
            deriv_data['signal1_deltas'][:] = [x / analysis_config['cable_length'] for x in deriv_data['signal1_deltas']]

        ax4.set_ylabel('$\\Delta$t [fs/m/K]')
        ax5.set_ylabel('$\\Delta$A [dB/m/K]')


    #######################################################################################################################
    # Closeout
    #######################################################################################################################
    #manual scaling if y axis. Auto is broken with scatter.
    ax4_mins = []
    ax4_maxs = []
    ax5_mins = []
    ax5_maxs = []

    for measurement_set in measurement_sets:
        ax4.scatter(derivatives[measurement_set]['x_deltas'], derivatives[measurement_set]['signal0_deltas'], marker='+')
        ax5.scatter(derivatives[measurement_set]['x_deltas'], derivatives[measurement_set]['signal1_deltas'], marker='+')
        if not len(derivatives[measurement_set]['signal0_deltas']) == 0:
            ax4_mins.append(min(derivatives[measurement_set]['signal0_deltas']))
            ax4_maxs.append(max(derivatives[measurement_set]['signal0_deltas']))
            ax5_mins.append(min(derivatives[measurement_set]['signal1_deltas']))
            ax5_maxs.append(max(derivatives[measurement_set]['signal1_deltas']))

    # set auto y range which does not work properly for small values in scatter
    if not len(ax4_maxs) == 0:
        dy4 = (max(ax4_maxs) - min(ax4_mins))*0.1
        ax4.set_ylim(min(ax4_mins)-dy4, max(ax4_maxs)+dy4)
        dy5 = (max(ax5_maxs) - min(ax5_mins))*0.1
        ax5.set_ylim(min(ax5_mins)-dy5, max(ax5_maxs)+dy5)

    fig.tight_layout()  # otherwise the legend is clipped

    fig.savefig(basename+measurement_set+'_analysis.pdf')
    
    plt.show()


if __name__ == '__main__':
    print('run \'./prototype.py -t first_tempsweep.txt -p -o tempsweep1\' to get the data needed for this plot.' )
    plot_sweep(np.arange(20., 31.+1.), [35.]*12, 'tempsweep1', 'temperature',
               ['1.3GHz', '1.0GHz', '3.0GHz', '6.0GHz', '10.0GHz'], ['S21_PHASE', 'S21_MAGNITUDE'],
               analysis_config={'type': 'rf_cable', 'normalise': [True, False], 'cable_length': 10,
                                'extra_signal_names': ['RF_FREQUENCY'], 'dut_name': 'simulated cable',
                                'time_string': '0:0:0'})
